/*
 * Map folding problem of The 6th International Olympiad in Informatics
 *
 * Written by:  Fredrik Hubinette (hubbe@lysator.liu.se)
 *
 */


/* This program tries every possible folding of the map that
 * doesn't bring two 'faces' of the map together that aren't listed
 * next to eachother in the input. It remembers what number is on
 * the upside and downside of each face, as the numbers in between
 * are of no consequence.
 *
 * The numbers remember are not the left-to-right numbering of the
 * faces, but rather the place this face should have in the final
 * stack of faces.
 */


#include <stdio.h>
#include <stdlib.h>

/* #define DEBUG */

/* every sqare, or 'face' of the map has one of these to remember
 * which number is on the upside, and which is on the downside.
 */
typedef struct PIECE
{
  int up,down;
} piece;


/* A whole map, I build new maps for each recursion, as I can't
 * 'undo' a folding easily with these data structures.
 */
typedef struct MAP
{
  int x,y;
  piece *map;
} map;

FILE *f;

/* The marco IND returns the piece on position (X,Y) in the map M */
#define IND(M,X,Y) ((M)->map[(Y)*((M)->x)+(X)])

/* this macro returns the largest of two numbers */
#define MAX(X,Y) ((X)>(Y)?(X):(Y))


/* Make a map of size x,y and initialize all pieces up and downsides
 * to -1 so that we know that they haven't been set yet.
 */
map *mkmap(int x,int y)
{
  map *m;
  m=(map *)malloc(sizeof(map));
  if(!m)
  {
    fprintf(stderr,"Out of memory.\n");
    exit(1);
  }
  m->map=(piece *)malloc(sizeof(piece)*x*y);
  if(!m->map)
  {
    fprintf(stderr,"Out of memory.\n");
    exit(1);
  }
  m->x=x;
  m->y=y;

  for(y=0;y<m->y;y++)
    for(x=0;x<m->x;x++)
      IND(m,x,y).up=IND(m,x,y).down=-1;
  return m;
}

/* Return a maps memory to the system */
void free_map(map *m)
{
  free(m->map);
  free(m);
}

#ifdef DEBUG
/* Print a map to stdout, for debug purposes */
void dumpmap(map *m)
{
  int x,y;
  for(y=0;y<m->y;y++)
  {
    for(x=0;x<m->x;x++)
    {
      printf("%3d/%-3d",IND(m,x,y).up,IND(m,x,y).down);
    }
    printf("\n");
  }
  getchar();
}
#endif


/* I need a linked list to remember the answers, 
 * as I need to print the answers backwards.
 */
typedef struct ANSWER
{
  struct ANSWER *next;
  char ans[20];
} answer;


/* Add a line to the linked list that make up the final answer */
answer *mkanswer(answer *tmp,char *msg,int w)
{
  answer *ret;
  ret=(answer *)malloc(sizeof(answer));
  if(!ret)
  {
    fprintf(stderr,"Out of memory.\n");
    exit(1);
  }

  ret->next=tmp;
  sprintf(ret->ans,msg,w);
  return ret;
}

/* Recure and do every folding that doesn't put faces together that
 * shouldn't be together in the final stack.
 */
answer *recur(map *m)
{
  int x,y,t,diff;
  answer *tmp;

#ifdef DEBUG
  dumpmap(m);
#endif

  /* have we found a winner? */
  if(m->x==1 && m->y==1 && IND(m,0,0).up<IND(m,0,0).down)
    return mkanswer(NULL,"\n",0);


  /* Try folding along the Y axis */
  for(x=1;x<m->x;x++)
  {
    for(y=0;y<m->y;y++)
    {
      for(t=0;t<x;t++)
      {
	if(x+t>=m->x) continue;
	diff=IND(m,x-t-1,y).up-IND(m,x+t,y).up;
#ifdef DEBUG
	printf("(%d,%d) - (%d,%d) = %d\n",x-t-1,y,x+t,y,diff);
#endif
	if(diff!=1 && diff!=-1) break;
      }
      if(t<x) break;
    }
    /* We have found a possible fold, let't try it, with
     * both the left and the right part
     */
    if(y>=m->y)
    {
      int xsize;
      map *newmap;

      xsize=MAX(m->x-x,x);
      newmap=mkmap(xsize,y);
      for(y=0;y<m->y;y++)
      {
	for(t=x;t<m->x;t++) IND(newmap,t-x,y)=IND(m,t,y);

	for(t=0;t<x;t++)
	{
	  IND(newmap,t,y).up=IND(m,x-t-1,y).down;
	  if(IND(newmap,t,y).down==-1)
	    IND(newmap,t,y).down=IND(m,x-t-1,y).up;
	}
      }
      tmp=recur(newmap);
      if(tmp)
      {
	free_map(m);
	return mkanswer(tmp,"V%dL ",x);
      }

      xsize=MAX(m->x-x,x);
      newmap=mkmap(xsize,y);
      for(y=0;y<m->y;y++)
      {
	for(t=0;t<x;t++) IND(newmap,xsize-x+t,y)=IND(m,t,y);

	for(t=x;t<m->x;t++)
	{
	  IND(newmap,xsize-(t-x)-1,y).up=IND(m,t,y).down;
	  if(IND(newmap,xsize-(t-x)-1,y).down==-1)
	    IND(newmap,xsize-(t-x)-1,y).down=IND(m,t,y).up;
	}
      }
      tmp=recur(newmap);
      if(tmp)
      {
	free_map(m);
	return mkanswer(tmp,"V%dR ",x);
      }
    }
  }


  /* Now we try to fold along the X axis */
  for(y=1;y<m->y;y++)
  {
    for(x=0;x<m->x;x++)
    {
      for(t=0;t<y;t++)
      {
	if(y+t>=m->y) continue;
	diff=IND(m,x,y-t-1).up-IND(m,x,y+t).up;
	if(diff!=1 && diff!=-1) break;
      }
      if(t<y) break;
    }
    /* Aha, we try to fold both upwards and downwards */
    if(x>=m->x)
    {
      int ysize;
      map *newmap;

      ysize=MAX(m->y-y,y);
      newmap=mkmap(x,ysize);
      for(x=0;x<m->x;x++)
      {
	for(t=y;t<m->y;t++) IND(newmap,x,t-y)=IND(m,x,t);

	for(t=0;t<y;t++)
	{
	  IND(newmap,x,t).up=IND(m,x,y-t-1).down;
	  if(IND(newmap,x,t).down==-1)
	    IND(newmap,x,t).down=IND(m,x,y-t-1).up;
	}

      }
      tmp=recur(newmap);
      if(tmp)
      {
	free_map(m);
	return mkanswer(tmp,"H%dU ",y);
      }

      ysize=MAX(m->y-y,y);
      newmap=mkmap(x,ysize);
      for(x=0;x<m->x;x++)
      {
	for(t=0;t<y;t++) IND(newmap,x,ysize-y+t)=IND(m,x,t);

	for(t=y;t<m->y;t++)
	{
	  IND(newmap,x,ysize-(t-y)-1).up=IND(m,x,t).down;
	  if(IND(newmap,x,ysize-(t-y)-1).down==-1)
	    IND(newmap,x,ysize-(t-y)-1).down=IND(m,x,t).up;
	}
      }

      tmp=recur(newmap);
      if(tmp)
      {
	free_map(m);
	return mkanswer(tmp,"H%dL ",y);
      }
    }
  }

  /* nothing interesting found, free the map
   * and return with dissappointment
   */
  free_map(m);
  return 0;
}


int main()
{
  answer *a;
  map *m;
  int i,x,y;

  f=fopen("input.txt","rt");
  if(!f)
  {
    perror("No such file");
    exit(1);
  }

  if(fscanf(f,"%d %d\n",&y,&x)!=2)
  {
    fprintf(stderr,"Error in indata.\n");
    exit(1);
  }

  m=mkmap(x,y);


  for(i=0;i<m->x*m->y;i++)
  {
    char num;

    if(!fscanf(f,"%c",&num))
    {
      fprintf(stderr,"Error in input file.\n");
      exit(1);
    }
    if(num<'A' || num>'Z')
    {
      fprintf(stderr,"Error in input file (%c).\n",num);
      exit(1);
    }
    num-='A';
    if(num<0 || num>=m->x*m->y)
    {
      fprintf(stderr,"Part number out of range %d (%d)\n",num,i+1);
      exit(1);
    }

    x=num % m->x;
    y=num / m->x;


    IND(m,x,y).up=IND(m,x,y).down=i;
  }

  fclose(f);

  /* Let's try folding it a few times..... */

  a=recur(m);


  /* print the answer - backwards */

  f=fopen("output.txt","w");
  if(!f)
  {
    fprintf(stderr,"Couldn't open output file.\n");
    exit(1);
  }

  if(!a)
  {
    fprintf(stderr,"Impossible.\n");
  }else{
    for(;a;a=a->next) fprintf(f,"%s",a->ans);
  }
  fclose(f);
  return 0;
}


