/*
 * clockwork.c by Lars Engebretsen
 *
 * Inputs: The states of the nine clocks as nine numbers between
 *         0 and 3, separated by a whitespace.
 *
 * Outputs: The moves that has to be made, sorted in ascending order.
 */

/* Examples: Input              Output
 *           2 1 3 0 0 2 2 2 0  13457
 *           1 0 0 0 0 0 0 0 0  1234566788
 *           0 1 2 0 0 1 2 3 0  1133344556667788999
 *           1 1 1 1 1 1 0 0 0  111223334455566777888999
 */

/*
 * We quickly realize that:
 * a) The order of the moves are irrelevant.
 * b) There can never be more than three moves of the same kind in the
 *    solution.
 *
 * If we number the clocks from 1 to 9 like this:
 *
 *   1 2 3
 *   4 5 6
 *   7 8 9
 *
 * and let n[i] be the number of times we make move i, and lets c[i] be
 * the initial state of clock i, it is obvious (at least after a moment
 * of thought) that n[] satisfies the following equation:
 *
 *   c[1]+n[1]+n[2]+n[4]           = 0 (mod 4)
 *   c[2]+n[1]+n[2]+n[3]+n[5]      = 0 (mod 4)
 *   c[3]+n[2]+n[3]+n[6]           = 0 (mod 4)
 *   c[4]+n[1]+n[4]+n[5]+n[7]      = 0 (mod 4)
 *   c[5]+n[1]+n[3]+n[5]+n[7]+n[9] = 0 (mod 4)
 *   c[6]+n[3]+n[5]+n[6]+n[9]      = 0 (mod 4)
 *   c[7]+n[4]+n[7]+n[8]           = 0 (mod 4)
 *   c[8]+n[5]+n[8]+n[8]+n[9]      = 0 (mod 4)
 *   c[9]+n[6]+n[9]+n[9]           = 0 (mod 4)
 *
 * or
 *
 *   c+A*n = 4*k      where k is a vector of nine arbitrary integers
 *
 * with
 *
 *         /1 1 0 1 0 0 0 0 0\
 *        | 1 1 1 0 1 0 0 0 0 |
 *        | 0 1 1 0 0 1 0 0 0 |
 *        | 1 0 0 1 1 0 1 0 0 |
 *   A == | 1 0 1 0 1 0 1 0 1 |
 *        | 0 0 1 0 1 1 0 0 1 |
 *        | 0 0 0 1 0 0 1 1 0 |
 *        | 0 0 0 0 1 0 1 1 1 |
 *         \0 0 0 0 0 1 0 1 1/
 *
 * It turns out that A has an inverse
 *
 *                /-1  2 -1  2  2 -3 -1 -3  4\
 *               |  3 -1  3 -1 -1 -1 -2  4 -2 |
 *               | -1  2 -1 -3  2  2  4 -3 -1 |
 *             1 |  3 -1 -2 -1 -1  4  3 -1 -2 |
 *   inv(A) == - | -1  2 -1  2 -3  2 -1  2 -1 |
 *             5 | -2 -1  3  4 -1 -1 -2 -1  3 |
 *               | -1 -3  4  2  2 -3 -1  2 -1 |
 *               | -2  4 -2 -1 -1 -1  3 -1  3 |
 *                \ 4 -3 -1 -3  2  2 -1  2 -1/
 *
 * and, as we can choose k[i] such that n[i]==0..3, there is always a
 * solution
 *
 *   n = inv(A)*(4*k-c)
 */

#include <stdio.h>

int c[9],n[9];

void main(int argc, char **argv)
{
    FILE *infile,*outfile;
    register int i,j;

    /* In C arrays always starts on 0. */

    /*
     * Read the initial positions of the clocks from input.txt
     */
    infile=fopen("input.txt","r");
    for(i=0; i<9; i++)
        fscanf(infile,"%d",&c[i]);
    fclose(infile);

    /*
     * We skip the 1/5 in inv(A) for now. By doing that we avoid
     * floating point arithmetics.
     */
    n[0]=   c[0]-2*c[1]+  c[2]-2*c[3]-2*c[4]+3*c[5]+  c[6]+3*c[7]-4*c[8];
    n[1]=-3*c[0]+  c[1]-3*c[2]+  c[3]+  c[4]+  c[5]+2*c[6]-4*c[7]+2*c[8];
    n[2]=   c[0]-2*c[1]+  c[2]+3*c[3]-2*c[4]-2*c[5]-4*c[6]+3*c[7]+  c[8];
    n[3]=-3*c[0]+  c[1]+2*c[2]+  c[3]+  c[4]-4*c[5]-3*c[6]+  c[7]+2*c[8];
    n[4]=   c[0]-2*c[1]+  c[2]-2*c[3]+3*c[4]-2*c[5]+  c[6]-2*c[7]+  c[8];
    n[5]= 2*c[0]+  c[1]-3*c[2]-4*c[3]+  c[4]+  c[5]+2*c[6]+  c[7]-3*c[8];
    n[6]=   c[0]+3*c[1]-4*c[2]-2*c[3]-2*c[4]+3*c[5]+  c[6]-2*c[7]+  c[8];
    n[7]= 2*c[0]-4*c[1]+2*c[2]+  c[3]+  c[4]+  c[5]-3*c[6]+  c[7]-3*c[8];
    n[8]=-4*c[0]+3*c[1]+  c[2]+3*c[3]-2*c[4]-2*c[5]+  c[6]-2*c[7]+  c[8];

    /*
     * Now it's time to choose k[] such that n becomes an integer (that
     * is n%5==0 right now) between 0 and 3.
     */

    for(i=0; i<9; i++)
    {
        while(n[i]%5)
            n[i]+=4;
        /*
         * Now we can take in account the 1/5 in inv(A).
         */
        n[i]/=5;
        /*
         * Now we make sure that n is in 0..3
         */
        n[i]%=4;
        while(n[i]<0)
            n[i]+=4;
    }

    /*
     * All we have to do now is to write the solution to output.txt
     */
    outfile=fopen("output.txt","w");
    for(i=0; i<9; i++)
        for(j=0; j<n[i]; j++)
            fprintf(outfile,"%d", i+1);
    fclose(outfile);
}
