/**************************************************************
*                                                             *
*          6th International Olympiad in Informatics          *
*                                                             *
*                         The castle                          *
*                                                             *
*                 Solution by Marcus Comstedt                 *
*                                                             *
**************************************************************/

#include <stdio.h>
#include <stdlib.h>

#define MAX_ROWS    (50)
#define MAX_COLUMNS (50)
#define MAX_CASTLE  (MAX_ROWS * MAX_COLUMNS)

int wall[MAX_ROWS][MAX_COLUMNS];
int equiv[MAX_CASTLE];
int room[MAX_CASTLE];

int castle_rows,castle_columns,castle_size;

FILE *out_file;


/* Get the wall data */

void read_input(void)
{
  FILE *f;
  int i,j;

  if(!(f = fopen("input.txt", "rt"))) {
	 fprintf(stderr, "Couldn't open INPUT.TXT!\n");
	 exit(10);
  }


  fscanf(f, "%d %d",&castle_rows,&castle_columns);

  /* Compute total number of modules in the castle */
  castle_size=castle_rows*castle_columns;

  /* Check dimensions for safety */
  if(castle_rows>MAX_ROWS || castle_columns>MAX_COLUMNS) {
	 fprintf(stderr,"Castle too large.\n");
	 exit(10);
  }

  /* And read the data for each module */
  for(i=0;i<castle_rows;i++)
	 for(j=0;j<castle_columns;j++)
		fscanf(f, "%d",&wall[i][j]);

  fclose(f);
}


/* Prepare output file */

void open_output()
{
  if(!(out_file = fopen("output.txt", "wt"))) {
	 fprintf(stderr, "Couldn't create OUTPUT.TXT!\n");
	 exit(10);
  }
}

/* Close output file */

void close_output()
{
  fclose(out_file);
}


/* Init data structures:  Make every module an equivalence class of its own,
	and init the room size counter for every such equivalence class to 0      */

void init_data(void)
{
  int i;

  for(i=0;i<castle_size;i++) {
	 equiv[i]=i;
	 room[i]=0;
  }
}


/* Find the root node of an equivalence class */

int find(int n)
{
  while(equiv[n]!=n) n=equiv[n];
  return n;
}


/* Merge two equivalence classes by putting the root node of class a
	under the root node of class b                                    */

void merge(int a, int b)
{
  equiv[find(a)]=find(b);
}


/* Iterate over the map, joining the equivalence classes of adjacent
	modules with no wall between them                                 */

void map_equiv(void)
{
  int i,j,n;

  for(i=0;i<castle_rows;i++)
	 for(j=0;j<castle_columns;j++) {

		/* Compute module number */
		n=i*castle_columns+j;

		/* Check wall to the north */
		if(!(wall[i][j]&2))
		  merge(n,n-castle_columns);

		/* And to the west */
		if(!(wall[i][j]&1))
		  merge(n,n-1);
	 }
}


/* Count the number of modules in each equivalence class (=room) and also
	make every node point directly to it's root node for improved speed in
	the next part of the program                                           */

void count_rooms(void)
{
  int i,n,rooms=0,largest=0;

  for(i=0;i<castle_size;i++) {

	 /* Find the module's root node */
	 n=find(i);

	 /* Compress path */
	 equiv[i]=n;

	 /* Increment number of modules in this room */
	 room[n]++;
  }

  for(i=0;i<castle_size;i++) {

	 /* If there is a roomsize for this module, it is a root node. Count it */
	 if(room[i]) rooms++;

	 /* Find largest roomsize too */
	 if(room[i]>largest) largest=room[i];
  }

  fprintf(out_file, "%d\n",rooms);
  fprintf(out_file, "%d\n",largest);
}


/* Try removing every wall in turn to see how large a room we get */

void remove_wall(void)
{
  int i,j,n,row,col,largest=0;
  char dir;

  for(i=0;i<castle_rows;i++)
	 for(j=0;j<castle_columns;j++) {

		/* Module number */
		n=i*castle_columns+j;

		/* Unless we are in the top row, try removing the north wall */
		if(i)

		  /* If the room to the north is equivalent, there is no wall */
	if(equiv[n]!=equiv[n-castle_columns])

			 /* Check the size of new room by adding sizes of both rooms */
			 if(room[equiv[n]]+room[equiv[n-castle_columns]]>largest) {

				/* It was larger.  Save size, position and direction */
				largest=room[equiv[n]]+room[equiv[n-castle_columns]];
				row=i;
				col=j;
				dir='N';
			 }

		/* Unless we are in the leftmost column, try removing the west wall */
		if(j)

		  /* If the room to the west is equivalent, there is no wall */
	if(equiv[n]!=equiv[n-1])

			 /* Check the size of new room by adding sizes of both rooms */
			 if(room[equiv[n]]+room[equiv[n-1]]>largest) {

				/* It was larger.  Save size, position and direction */
				largest=room[equiv[n]]+room[equiv[n-1]];
				row=i;
				col=j;
				dir='W';
			 }
	 }

  if(largest)
	 fprintf(out_file, "%d %d %c\n", row+1, col+1, dir);
  else {
	 fprintf(stderr, "No walls to remove.\n");
	 exit(10);
  }
}


void main()
{
  read_input();
  open_output();
  init_data();
  map_equiv();
  count_rooms();
  remove_wall();
  close_output();
  exit(0);
}
