/**************************************************************
*                                                             *
*          6th International Olympiad in Informatics          *
*                                                             *
*                 Houses                                      *
*                                                             *
*                 Solution by Marcus Comstedt                 *
*                                                             *
**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define MAX_HOUSES 30

#define MAX_CORNERS (MAX_HOUSES*2+2)

#define START 0
#define GOAL  1

#define FUZZYFACTOR 1e-7

/* Positions of the houses */

int house_x[MAX_HOUSES],house_y[MAX_HOUSES];

/* Coordinates of the houses northwest and southeast corners, the
   relative distance between them, and the next corner in the path. */

int corner_x[MAX_CORNERS],corner_y[MAX_CORNERS];
double distance[MAX_CORNERS][MAX_CORNERS];
int plink[MAX_CORNERS];

int num_houses,num_corners;


/* Add a corner to the database after checking that it is inside
   the area (0,0)-(100,100) and is not a dupe.                   */

void add_corner(int x, int y)
{
  int i;

  if(x>=0 && y>=0 && x<=100 && y<=100) {
    for(i=0;i<num_corners;i++)
      if(x==corner_x[i] && y==corner_y[i])
	return;
    corner_x[num_corners]=x;
    corner_y[num_corners]=y;
    num_corners++;
  }
}


/* Read map from file */

void read_input(void)
{
  int n,x,y;
  FILE *f;

  if(!(f=fopen("input.txt", "rb"))) {
	 fprintf(stderr, "Couldn't open INPUT.TXT!\n");
	 exit(10);
  }

  num_houses=0;
  num_corners=0;

  add_corner(0,0);      /* The startpoint, must be added first to become #0 */
  add_corner(100,100);  /* The goalpoint, must be added next to become #1 */

  fscanf(f, "%d",&n);
  while(n--) {
	 fscanf(f, "%d %d",&x,&y);

	 /* Skip houses completely outside the area (0,0)-(100,100) */
	 if(x<=-5 || y<=-5 || x>=100 || y>=100)
		continue;

    if(num_houses>=MAX_HOUSES) {
      fprintf(stderr,"Too many houses!  I will only let you have %d.\n",MAX_HOUSES);
      exit(10);
    }

    house_x[num_houses]=x;
    house_y[num_houses]=y;
    num_houses++;

    add_corner(x,y+5); /* northwest corner */
    add_corner(x+5,y); /* southeast corner */
  }

  fclose(f);
}


/* Determine on which side of a line a point is, return 1 or 2 */
/* or 3 if the point is exactly on the line                    */

int side(int x, int y, int x1, int y1, int x2, int y2)
{
  long l,p,s;

  /* Translate coordinates for easy calculations */

  x-=x1;
  y-=y1;
  x2-=x1;
  y2-=y1;

  /* Compute the scalar product of the normal of the line
     and the orthogonal projection of the point on the line... 8-/ */

  l=x2*x2+y2*y2;
  p=x*x2+y*y2;
  s=y2*(x*l-x2*p)-x2*(y*l-y2*p);

  return (s? (s<0? 1 : 2) : 3);  /* Return result depending on sign of s */
}


/* See if a straight path is clear from houses */

int clear(int x1, int y1, int x2, int y2)
{
  int i;

  for(i=0;i<num_houses;i++)
    if(house_x[i]+5>x1 && house_x[i]<x2 && house_y[i]+5>y1 && house_y[i]<y2)

      /* If a house has corners on both sides of the path, it is blocking. */

      if(!(side(house_x[i],house_y[i],x1,y1,x2,y2)
	  &side(house_x[i],house_y[i]+5,x1,y1,x2,y2)
	  &side(house_x[i]+5,house_y[i],x1,y1,x2,y2)
	  &side(house_x[i]+5,house_y[i]+5,x1,y1,x2,y2)))
	    return(0);
  return 1;
}


/* Find out the relative distances between all corners, and also between
   whice travel is possible at all                                       */

#define dist(x1,y1,x2,y2) (sqrt((double)((x1-x2)*(x1-x2)+(y1-y2)*(y1-y2))))

void calculate_distances(void)
{
  int i,j;

  for(i=0;i<num_corners;i++) {
	 plink[i]=GOAL;
	 for(j=0;j<num_corners;j++) {

		/* Mark the distance i to j as not passable */
		distance[i][j]= -1.0;

		/* Never travel south or west */
		if(corner_x[j]>=corner_x[i] && corner_y[j]>=corner_y[i] && i!=j)

		  /* Check that no house is blocking our path */
		  if(clear(corner_x[i],corner_y[i],corner_x[j],corner_y[j]))

			 /* Compute and store actual distance */
			 distance[i][j]=dist(corner_x[i],corner_y[i],corner_x[j],corner_y[j]);
	 }
  }
}


/* Find shortest path recursively */

double shortest_path(int from)
{
  int i,clink;
  double shortest,link;

  /* See if straight line possible, or minimal distance already calculated */
  if((shortest=distance[from][GOAL])>=0.0)
	 return shortest;

  /* Try rounding all corners in order */
  for(i=0;i<num_corners;i++)
	 if(distance[from][i]>=0.0)
		if((link=shortest_path(i))>=0.0)
		  if((link+=distance[from][i])<shortest || shortest<0.0) {
			 shortest=link;
			 clink=i;
		  }

  /* Save this result for future use also */
  plink[from]=clink;
  return distance[from][GOAL]=shortest;
}

void save_solution()
{
  FILE *f;

  int i=START;

  if(!(f=fopen("output.txt", "wt"))) {
	 fprintf(stderr, "Can't create OUTPUT.TXT!\n");
	 exit(10);
  }
  for(;;) {
	 fprintf(f, "%d %d\n", corner_x[i], corner_y[i]);
	 if(i==GOAL)
		break;
	 i=plink[i];
  }
  fclose(f);
}

void main()
{
  read_input();
  calculate_distances();
  if(shortest_path(START)<0.0)
	 printf("No solution found!?!  Did you place a house over the start- or endpoint?\n");
  else
	 save_solution();
}
